##############################################################################
#
# Copyright (c) 2002 Nexedi SARL and Contributors. All Rights Reserved.
#          Jean-Paul Smets-Solane <jp@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from AccessControl import ClassSecurityInfo
from Acquisition import aq_base

from Products.ERP5Type.Base import Base
from Products.ERP5Type.CopySupport import CopyContainer
from Products.ERP5Type import PropertySheet, Permissions

from Products.BTreeFolder2.CMFBTreeFolder import CMFBTreeFolder

import threading

from zLOG import LOG

class Folder( CopyContainer, CMFBTreeFolder, Base):
  """
  A Folder is a subclass of Base but not of XMLObject.
  Folders are not considered as documents and are therefore
  not synchronisable.

  ERP5 folders are implemented as CMFBTreeFolder objects
  and can store up to a million documents on a standard
  computer.
  ERP5 folders will eventually use in the near future the
  AdaptableStorage implementation in order to reach performances
  of 10 or 100 millions of documents in a single folder.

  ERP5 folders include an automatic id generation feature
  which allows user not to define an id when they create
  a new document in a folder.

  ERP5 folders use the ZSQLCatalog to search for objects
  or display content. This requires a method called
  *z_search_folder* to be put inside the ZSQLCatalog object
  of the ERP5 portal.

  An ERP5 Binder document class will eventually be defined
  in order to implement a binder of documents which can itself
  be categorized.
  """

  meta_type = 'ERP5 Folder'
  portal_type = 'Folder'
  add_permission = Permissions.AddPortalContent
  isPortalContent = 1
  isRADContent = 1

  # Declarative security
  security = ClassSecurityInfo()
  security.declareObjectProtected(Permissions.View)

  # Declarative properties
  property_sheets = ( PropertySheet.Base
                    , PropertySheet.XMLObject
                    )

  # Default Variables
  last_id = 0

  # CMF Factory Type Information
  factory_type_information = \
         { 'id'           : portal_type
         , 'meta_type'    : meta_type
         , 'description'  : """\
Folders allow to store a large number of documents (1,000,000 should not
be a problem)."""
         , 'icon'     : 'folder_icon.gif'
         , 'product'  : 'ERP5Type'
         , 'factory'  : 'addFolder'
         , 'filter_content_types' : 0
         , 'immediate_view' : 'folder_view'
         , 'actions'  :
        ( { 'id'    : 'view'
          , 'name'    : 'View'
          , 'action'  : 'folder_list'
          , 'permissions'   : (Permissions.View,)
          , 'category'  : 'folder'
          }
        , { 'id'    : 'edit'
          , 'name'    : 'Edit'
          , 'action'  : 'folder_edit_form'
          , 'permissions'   : (Permissions.ManageProperties,)
          , 'category'  : 'folder'
          }
        , { 'id'    : 'localroles'
          , 'name'    : 'Local Roles'
          , 'action'  : 'folder_localrole_form'
          , 'permissions'   :  (Permissions.ManageProperties,)
          , 'category'  : 'folder'
          }
        , { 'id'    : 'syndication'
          , 'name'    : 'Syndication'
          , 'action'  : 'synPropertiesForm'
          , 'permissions'   : (Permissions.ManageProperties,)
          , 'category'  : 'folder'
          }
        , { 'id'    : 'foldercontents'
          , 'name'    : 'Folder contents'
          , 'action'  : 'folder_contents'
          , 'permissions'   : (Permissions.ListFolderContents,)
          , 'category'  : 'folder'
          }
        )
        }

  # Class inheritance fixes
  edit = Base.edit
  _edit = Base._edit

  # Implementation
  security.declarePrivate('_setObject')
  def _setObject(self, id, object, roles=None, user=None, set_owner=1):
    """
      This method is here in order to dynamically update old
      folders into the new BTree folder type.
      This method is destructive in the sens that objects
      of the old folder will be lost during the update
    """
    # First make sur the folder has been initialized
    if not hasattr(self, '_tree'):
      CMFBTreeFolder.__init__(self, self.id)
    if not self._tree:
      CMFBTreeFolder.__init__(self, self.id)
    # Then insert the object
    CMFBTreeFolder._setObject(self, id, object, roles=roles, user=user, set_owner=set_owner)

  # Automatic ID Generation method
  security.declareProtected(Permissions.View, 'generateNewId')
  def generateNewId(self):
      """
        Generate a new Id which has not been taken yet in this folder.
        Eventually increment the id number until an available id
        can be found

        Permission is view because we may want to add content to a folder
        without changing the folder content itself. XXXXXXXXXXX
        XXXXXXXXXXXX
      """
      my_id = self.last_id
      l = threading.Lock()
      l.acquire()
      try:
        while self.hasObject(str(my_id)):
          self.last_id = self.last_id + 1
          my_id = self.last_id
      finally:
        l.release()
      return str(my_id)

  # Get the content
  security.declareProtected(Permissions.View, 'searchFolder')
  def searchFolder(self, **kw):
    """
      Search the content of a folder by calling
      the portal_catalog.
    """
    if not kw.has_key('parent_uid'): #WHY ????
      kw['parent_uid'] = self.uid
    kw2 = {}
    # Remove useless matter before calling the
    # catalog. In particular, consider empty
    # strings as None values
    for cname in kw.keys():
      if kw[cname] != '' and kw[cname]!=None:
        kw2[cname] = kw[cname]
    # The method to call to search the folder
    # content has to be called z_search_folder
    method = self.portal_catalog.portal_catalog
    return method(**kw2)

  # Count objects in the folder
  security.declarePrivate('_count')
  def _count(self, **kw):
    """
      Returns the number of items in the folder.
    """
    # PERFORMANCE PROBLEM
    # This should be improved in order to use
    # SQL counting
    return len(self.searchFolder(**kw))

  def updateAll(self, filter=None, method=None, test_after=None):
    """
    update all objects inside this particular folder wich
    returns not None to the test.

    filter have to be a method with one parameter (the object)
    wich returns None if we must not update the object

    test_after have to be a method with one parameter (the object)
    wich returns a string

    method is the update method with also one parameter

    """
    update_list = []
    for o in self.objectValues():
      # Test if we must apply the upgrade
      if filter(o):
        method(o)
        update_list += test_after(o)
      obase = aq_base(o)
      if hasattr(obase, 'updateAll'):
        update_list += o.updateAll(filter=filter, \
                              method=method, test_after=test_after)

    return update_list

  def upgradeObjectClass(self, test_before=None, from_class=None,\
                         to_class=None, test_after=None):
    """
    upgrade the class of all objects inside this
    particular folder
     test have to be a method with one parameter
     migrations is a dictionnary of class, { from_class : to_class }
    """
    LOG("upradeObjectClass: folder ",0,self.id)
    test_list = []
    folder = self.getObject()
    for o in self.listFolderContents():
      # Make sure this sub object is not the same as object
      if o.getPhysicalPath() != self.getPhysicalPath():
        id = o.getId()
        obase = aq_base(o)
        # Check if the subobject have to also be upgraded
        if hasattr(obase,'upgradeObjectClass'):
          test_list += o.upgradeObjectClass(test_before=test_before, \
                          from_class=from_class, to_class=to_class,
                          test_after=test_after)
        # Test if we must apply the upgrade
        if test_before(o) is not None:
          LOG("upradeObjectClass: id ",0,id)
          klass = obase.__class__
          LOG("upradeObjectClass: klass ",0,str(klass))
          LOG("upradeObjectClass: from_class ",0,str(from_class))
          if klass == from_class:
            try:
                newob = to_class(obase.id)
                newob.id = obase.id # This line activates obase.
            except:
                newob = to_class(id)
                newob.id = id
            keys = obase.__dict__.keys()
            for k in keys:
              if k not in ('id', 'meta_type', '__class__'):
                setattr(newob,k,obase.__dict__[k])

            self.manage_delObjects(id)
            LOG("upradeObjectClass: ",0,"add new object: %s" % str(newob.id))
            get_transaction().commit() # XXX this commit should be after _setObject
            LOG("upradeObjectClass: ",0,"newob.__class__: %s" % str(newob.__class__))
            self._setObject(id, newob)
            object_to_test = self._getOb(id)
            test_list += test_after(object_to_test)

    return test_list


  # Catalog related
  security.declareProtected( Permissions.ModifyPortalContent, 'reindexObject' )
  def reindexObject(self, idxs=[]):
    """
      Fixes the hierarchy structure (use of Base class)
      XXXXXXXXXXXXXXXXXXXXXXXX
      BUG here : when creating a new base category
    """
    return Base.reindexObject(self)

  # Special Relation keyword : 'content' and 'container'
  security.declareProtected( Permissions.AccessContentsInformation, '_getCategoryMembershipList' )
  def _getCategoryMembershipList(self, category,
                          spec=(), filter=filter, portal_type=portal_type, base=0 ):
    if category == 'content':
      content_list = self.searchFolder(portal_type=spec)
      return map(lambda x: x.relative_url, content_list)
    else:
      return Base.getCategoryMembershipList(self, category,
          spec=spec, filter=filter, portal_type=portal_type,  base=base)

  # Alias - class inheritance resolutino
  security.declareProtected( Permissions.View, 'Title' )
  Title = Base.Title

  security.declareProtected(Permissions.AccessContentsInformation, 'checkConsistency')
  def checkConsistency(self, fixit=0):
    """
    Check the consistency of this object, then
    check recursively the consistency of every sub object.

    """
    error_list = []
    # Fix BTree
    if fixit:
      btree_ok = self._cleanup()
      if not btree_ok:
        get_transaction().commit() # We must commit if we want to keep on recursing
        error_list += [(self.getRelativeUrl(), 'BTree Inconsistency', 199, '(fixed)')]
    # Call superclass
    error_list += Base.checkConsistency(self, fixit=fixit)
    if fixit: get_transaction().commit() # We must commit before listing folder contents
                                         # in case we erased some data
    # Then check the consistency on all sub objects
    for object in self.contentValues():
      if fixit:
        extra_errors = object.fixConsistency()
      else:
        extra_errors = object.checkConsistency()
      if len(extra_errors) > 0:
        error_list += extra_errors
        # Commit after each subobject
        #if fixit:
        try:
          get_transaction().commit()
        except:
          LOG("Folder WARNING",0,
            "Could not commit checkConsistency transaction for object %s" % object.getRelativeUrl())

    # We should also return an error if any
    return error_list

  # Multiple Inheritance Priority Resolution
  _setProperty = Base._setProperty
  setProperty = Base.setProperty
  getProperty = Base.getProperty
  hasProperty = Base.hasProperty

  # Aliases
  getObjectIds = CMFBTreeFolder.objectIds
